package io.gitlab.jfronny.inceptum.common.api;

import io.gitlab.jfronny.commons.HttpUtils;
import io.gitlab.jfronny.inceptum.common.Net;
import io.gitlab.jfronny.inceptum.common.Utils;
import io.gitlab.jfronny.inceptum.common.model.maven.*;
import org.jetbrains.annotations.Nullable;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

import javax.xml.parsers.*;
import javax.xml.stream.XMLStreamException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URISyntaxException;
import java.nio.file.Path;
import java.util.*;

public class MavenApi {
    private static final DocumentBuilder FACTORY;
    private static final Set<String> RUNTIME_SCOPES = Set.of("compile", "runtime");

    static {
        try {
            FACTORY = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new RuntimeException("Could not create document builder", e);
        }
    }

    public static Path downloadLibrary(String repo, ArtifactMeta meta) throws IOException, URISyntaxException {
        Path res = meta.getLocalPath();
        Net.downloadFile(Utils.join("/", repo, meta.getJarPath(true)), res);
        return res;
    }

    public static Pom getPom(String repo, ArtifactMeta meta) throws IOException, SAXException, URISyntaxException, XMLStreamException {
        try (InputStream is = HttpUtils.get(Utils.join("/", repo, meta.getPomPath())).sendInputStream()) {
            Document doc = FACTORY.parse(is);
            doc.getDocumentElement().normalize();
            Pom result = new Pom();
            if (!"project".equals(doc.getDocumentElement().getNodeName())) throw new IOException("Illegal document name");
            boolean hasModelVersion = false;
            boolean hasGroupId = false;
            boolean hasArtifactId = false;
            boolean hasVersion = false;
            for (Node node : iterable(doc.getDocumentElement().getChildNodes())) {
                switch (node.getNodeName()) {
                    case "modelVersion" -> {
                        hasModelVersion = true;
                        result.modelVersion = node.getTextContent();
                    }
                    case "parent" -> {
                        // Dirty hack to get slf4j working: simply assume the groupId and version of the parent is also the groupId of this
                        if (!hasGroupId) {
                            for (Node child : iterable(node.getChildNodes())) {
                                switch (child.getNodeName()) {
                                    case "groupId" -> {
                                        if (!hasGroupId) {
                                            hasGroupId = true;
                                            result.groupId = node.getTextContent();
                                        }
                                    }
                                    case "version" -> {
                                        if (!hasVersion) {
                                            hasVersion = true;
                                            result.version = node.getTextContent();
                                        }
                                    }
                                }
                            }
                        }
                    }
                    case "groupId" -> {
                        hasGroupId = true;
                        result.groupId = node.getTextContent();
                    }
                    case "artifactId" -> {
                        hasArtifactId = true;
                        result.artifactId = node.getTextContent();
                    }
                    case "version" -> {
                        hasVersion = true;
                        result.version = node.getTextContent();
                    }
                    case "packaging" -> result.packaging = node.getTextContent();
                    case "dependencies" -> {
                        result.dependencies = new LinkedList<>();
                        for (Node dep : iterable(node.getChildNodes())) {
                            MavenDependency resolved = parseDependency(dep);
                            if (resolved != null) {
                                result.dependencies.add(resolved);
                            }
                        }
                    }
                    case "classifier" -> result.classifier = node.getTextContent();
                    default -> {}
                }
            }
            if (!hasModelVersion) throw new IOException("Pom lacks modelVersion");
            if (!hasGroupId) throw new IOException("Pom lacks groupId");
            if (!hasArtifactId) throw new IOException("Pom lacks artifactId");
            if (!hasVersion) throw new IOException("Pom lacks version");
            return result;
        }
    }

    private static @Nullable MavenDependency parseDependency(Node doc) throws IOException {
        MavenDependency result = new MavenDependency();
        boolean hasGroupId = false;
        boolean hasArtifactId = false;
        boolean hasVersion = false;
        boolean hasScope = false;
        for (Node node : iterable(doc.getChildNodes())) {
            switch (node.getNodeName()) {
                case "groupId" -> {
                    hasGroupId = true;
                    result.groupId = node.getTextContent();
                }
                case "artifactId" -> {
                    hasArtifactId = true;
                    result.artifactId = node.getTextContent();
                }
                case "version" -> {
                    hasVersion = true;
                    result.version = node.getTextContent();
                }
                case "scope" -> {
                    hasScope = true;
                    result.scope = node.getTextContent();
                    if (!RUNTIME_SCOPES.contains(result.scope)) return null;
                }
                case "optional" -> {
                    if (node.getTextContent().equals("true")) return null;
                }
            }
        }
        if (!hasGroupId) throw new IOException("Pom lacks groupId");
        if (!hasArtifactId) throw new IOException("Pom lacks artifactId");
        if (!hasVersion) {
            if (result.groupId.equals("org.lwjgl")) {
                // Lwjgl uses a shared bom for versions which I don't want to support
                // The required modules are explicit dependencies of launcher-imgui anyway
                return null;
            }
            throw new IOException("Dependency " + result.groupId + ":" + result.artifactId + " lacks version");
        }
        if (!hasScope) throw new IOException("Pom lacks scope");
        return result;
    }

    public static ArtifactMeta getMetadata(String repo, String artifact) throws IOException, SAXException, URISyntaxException {
        ArtifactMeta sourceMeta = ArtifactMeta.parse(artifact);
        try (InputStream is = HttpUtils.get(Utils.join("/", repo, sourceMeta.getMetadataPath())).sendInputStream()) {
            Document doc = FACTORY.parse(is);
            doc.getDocumentElement().normalize();
            ArtifactMeta result = new ArtifactMeta();
            if (!"metadata".equals(doc.getDocumentElement().getNodeName())) throw new IOException("Illegal document name");
            boolean hasGroupId = false;
            boolean hasArtifactId = false;
            boolean hasVersion = false;
            for (Node node : iterable(doc.getDocumentElement().getChildNodes())) {
                switch (node.getNodeName()) {
                    case "groupId" -> {
                        hasGroupId = true;
                        result.groupId = node.getTextContent();
                    }
                    case "artifactId" -> {
                        hasArtifactId = true;
                        result.artifactId = node.getTextContent();
                    }
                    case "version" -> {
                        hasVersion = true;
                        result.version = node.getTextContent();
                    }
                    case "versioning" -> {
                        for (Node node1 : iterable(node.getChildNodes())) {
                            if (node1.getNodeName().equals("snapshot")) {
                                String timestamp = null;
                                String buildNumber = null;
                                for (Node node2 : iterable(node1.getChildNodes())) {
                                    switch (node2.getNodeName()) {
                                        case "timestamp" -> timestamp = node2.getTextContent();
                                        case "buildNumber" -> buildNumber = node2.getTextContent();
                                        default -> {}
                                    }
                                }
                                if (timestamp == null) throw new IOException("Pom snapshots lack timestamp");
                                if (buildNumber == null) throw new IOException("Pom snapshots lack buildNumber");
                                result.snapshotVersion = timestamp + '-' + buildNumber;
                            }
                        }
                    }
                    default -> {}
                }
            }
            if (!hasGroupId) throw new IOException("Pom lacks groupId");
            if (!hasArtifactId) throw new IOException("Pom lacks artifactId");
            if (!hasVersion) throw new IOException("Pom lacks version");
            result.classifier = sourceMeta.classifier;
            return result;
        }
    }

    private static Iterable<Node> iterable(NodeList list) {
        return () -> new Iterator<>() {
            int index = 0;

            @Override
            public boolean hasNext() {
                while (index < list.getLength() && isWhitespace(list.item(index))) {
                    index++;
                }
                return index < list.getLength();
            }

            @Override
            public Node next() {
                if (!hasNext()) throw new NoSuchElementException();
                return list.item(index++);
            }
        };
    }

    private static boolean isWhitespace(Node node) {
        if (node.getNodeType() == Node.TEXT_NODE && node.getTextContent().isBlank()) return true;
        if (node.getNodeType() == Node.COMMENT_NODE) return true;
        return false;
    }
}
